{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Objektumorientált programozás II."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "<dl><dt> Egységbezárás (encapsulation)</dt><dd> a procedurális megközelítéssel ellentétben az adatok és a függvények a program nem különálló részét képezik, azok összetartoznak: együtt alkotnak egy objektumot. Az objektum felülete(ke)n (interfész, interface) keresztül érhető el a többi objektum számára.\n",
    "</dd></dl>\n",
    "<dl><dt> Osztály (class)</dt><dd> egy objektum prototípusa, tulajdonságokkal ellátva, melyek jellemzik az osztály példányait. A tulajdonságok osztály és példányváltozók, és metódusok (tagfüggvények). Pont-jelöléssel érhetők el.\n",
    "</dd></dl>\n",
    "<dl><dt> Attribútum/tulajdonság (attribute)</dt><dd> egy objektum tulajdonsága, jellemzője, az objektum nevét követő pont után írjuk a nevét.\n",
    "</dd></dl>\n",
    "<dl><dt> Példány (instance)</dt><dd> egy osztályhoz tartozó egyedi objektum. \n",
    "</dd></dl>\n",
    "<dl><dt> Példányosítás (instantiation)</dt><dd> egy osztály egy példányának létrehozása.\n",
    "</dd></dl>\n",
    "<dl><dt> Példány változó (instance variable)</dt><dd> metóduson belül definiált változó, mely csak az osztály adott példányához tartozik.\n",
    "</dd></dl>\n",
    "<dl><dt> Osztályváltozó (class variable)</dt><dd> változó, melyet az osztály minden példánya elér.\n",
    "</dd></dl>\n",
    "<dl><dt> Metódus (method, tagfüggvény)</dt><dd> osztályon belül definiált spec. függvény, első argumentuma tipikusan a self.\n",
    "</dd></dl>\n",
    "<dl><dt> Öröklődés (inheritance)</dt><dd> származtatással egy már meglévő osztályból egy újat hozunk létre. Ez rendelkezni fog az ős minden tulajdonságával, amihez továbbiakat ad(hat)unk.\n",
    "</dd></dl>\n",
    "<dl><dt> Polimorfizmus/többalakúság (polymorphism)</dt><dd> különböző viselkedésmódokkal ruházzuk fel az egymásból származtatott objektumokat.\n",
    "</dd></dl>\n",
    "<dl><dt> Névtér (namespace)</dt><dd> megadja, hogy egy név melyik objektumhoz tartozik (leképezés a nevekről az objektumokra). (Példák névterekre: foglalt nevek, globális változók, egy függvény lokális nevei).\n",
    "</dd></dl>\n",
    "<dl><dt> Hatókör/szkóp (scope)</dt><dd> a kód azon része, ahol a névtér közvetlenül (a teljes elérési út magadása nélkül) elérhető.\n",
    "</dd></dl>\n",
    "\n"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "## Öröklődés\n",
    "Írunk egy Person osztályt. Minden embernek lesz neve és titulusa, a `__repr__` függvénnyel pedig megszólítjuk. A Knight osztály a Person leszármazottja. Minden lovag egyben ember is, de a megszólításuk egyedi. Leszármaztatni akkor tudunk, ha az ősosztályunk is az object leszármazottja. A leszármazott örökli az ősosztály összes tagváltozóját és metódusát."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Person(object):\n",
    "    def __init__(self, name, title):\n",
    "        self.name = name\n",
    "        self.title = title\n",
    "    def __repr__(self):\n",
    "        return self.title + \" \" + self.name\n",
    "\n",
    "class Knight(Person):\n",
    "    def __init__(self, name):\n",
    "        super(Knight, self).__init__(name, 'Sir')\n",
    "\n",
    "varga = Person('Varga', 'Mr')\n",
    "launcelot = Knight('Launcelot')\n",
    "print varga\n",
    "print launcelot"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "A leszármazott osztályban definiálhatunk új tagváltozókat is. Például legyen minden lovagnak egy *epitheton ornansa*!"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Person(object):\n",
    "    def __init__(self, name, title):\n",
    "        self.name = name\n",
    "        self.title = title\n",
    "    def __repr__(self):\n",
    "        return self.title + \" \" + self.name\n",
    "\n",
    "class Knight(Person):\n",
    "    def __init__(self, name, eo):\n",
    "        super(Knight, self).__init__(name, 'Sir')\n",
    "        self.eo = eo\n",
    "         \n",
    "launcelot = Knight('Launcelot', 'the brave')\n",
    "launcelot"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Minden lovag megérdemli, hogy a neve mellé az állandó jezőjét is hozzáfűzzük. Ehhez átdefiniáljuk a `__repr__` metódust. Azonos nevű metódusok esetén mindig a leszármazottban definiált élvez elsőbbséget és hívódik meg."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Person(object):\n",
    "    def __init__(self, name, title):\n",
    "        self.name = name\n",
    "        self.title = title\n",
    "    def __repr__(self):\n",
    "        return self.title + \" \" + self.name\n",
    "\n",
    "class Knight(Person):\n",
    "    def __init__(self, name, eo):\n",
    "        super(Knight, self).__init__(name, 'Sir')\n",
    "        self.eo = eo\n",
    "    def __repr__(self):\n",
    "        return self.title + \" \" + self.name + \", \" + self.eo\n",
    "         \n",
    "launcelot = Knight('Launcelot', 'the brave')\n",
    "robin = Knight('Robin', 'the Not-quite-so-brave-as-Sir-Launcelot')\n",
    "print launcelot\n",
    "print robin"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "## Példányváltozók, osztályváltozók\n",
    "Van mód arra is, hogy bizonyos változókat osztályszinten definiáljunk. Ilyen például a lovagok \"Sir\" megszólítása. Az osztályszinten definiált title azonban egybeesik a Person title változójával így a lovagok esetében a `self.title` erre a változóra mutat. Ahhoz, hogy az osztályszintű változót elérjük, a `Knight.title`-ra kell hivatkozni. A `searching_for` azonban már úgy viselkedik, ahogyan szeretnénk, ez osztályváltozó."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Person(object):\n",
    "    def __init__(self, name, title=\"\"):\n",
    "        self.name = name\n",
    "        self.title = title\n",
    "    def __repr__(self):\n",
    "        return self.title + \" \" + self.name\n",
    "\n",
    "class Knight(Person):\n",
    "    title = 'Sir'\n",
    "    searching_for = 'The Holy Grail'\n",
    "    def __init__(self, name, eo):\n",
    "        super(Knight, self).__init__(name)\n",
    "        self.eo = eo\n",
    "    def __repr__(self):\n",
    "        return Knight.title + \" \" + self.name + \", \" + self.eo\n",
    "         \n",
    "launcelot = Knight('Launcelot', 'the brave')\n",
    "robin = Knight('Robin', 'the Not-quite-so-brave-as-Sir-Launcelot')\n",
    "print launcelot\n",
    "print robin\n",
    "print launcelot.searching_for\n",
    "print robin.searching_for"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false
   },
   "outputs": [],
   "source": [
    "print launcelot.title, \"-\", Knight.title\n",
    "print launcelot.searching_for, \"-\", Knight.searching_for"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Megjegyezzük, hogy az öröklési lánc tetszőlegesen hosszú lehet, illetve lehet egyszerre több osztálytól örökölni."
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "## Iterálható objektumok\n",
    "Láttuk, hogy a `for i in L` nem csak akkor működik, ha `L` lista. A `for` meghívja az `iter()` függvényt, mely egy iterálható objektumot ad vissza, melyre definiálva van egy `next()` függvény, amely minden meghívására visszaadja az objektum egy elemét. Ha a `next()` nem talál több elemet, egy `StopIteration` kivételt dob."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "r = range(3)\n",
    "it = iter(r)\n",
    "it.next()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "print it.next()\n",
    "print it.next()"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "it.next()"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Tetszőleges, általunk definiált osztályhoz is adható iterátor tulajdonság. Először is szükség van egy `__iter__` függvényre, ami egy olyan objektummal tér vissza, ami iterálható. Az iterálható objekcumok (`list, set, tuple`) definiálnak egy `next()` metódust, ami sorban visszaadja az elemeket."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Person(object):\n",
    "    def __init__(self, name, title=\"\"):\n",
    "        self.name = name\n",
    "        self.title = title\n",
    "    def __repr__(self):\n",
    "        return self.title + \" \" + self.name\n",
    "\n",
    "class Knight(Person):\n",
    "    title = 'Sir'\n",
    "    searching_for = 'The Holy Grail'\n",
    "    def __init__(self, name, eo):\n",
    "        super(Knight, self).__init__(name)\n",
    "        self.eo = eo\n",
    "    def __repr__(self):\n",
    "        return Knight.title + \" \" + self.name + \", \" + self.eo    \n",
    "\n",
    "class Group(object):\n",
    "    def __init__(self, name, persons):\n",
    "        self.persons = persons\n",
    "        self.name = name\n",
    "        self.index = -1\n",
    "    def __iter__(self):\n",
    "        return self\n",
    "    def next(self):\n",
    "        if self.index == len(self.persons) - 1:\n",
    "            self.index = -1\n",
    "            raise StopIteration     # dobunk egy kivetelt\n",
    "        self.index += 1\n",
    "        return self.persons[self.index]\n",
    "        \n",
    "kotrt = Group('Knights of The Round Table', \n",
    "              [Knight('Launcelot', 'the brave'), \n",
    "               Knight('Galahad', 'the pure'),\n",
    "               Knight('Bedevere', 'the wise'), \n",
    "               Knight('Robin', 'the Not-quite-so-brave-as-Sir-Launcelot')])\n",
    "for knight in kotrt:\n",
    "    print knight"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Vigyázzunk, hogy az indexet vissza kell állítani, különben csak egyszer tudunk iterálni! Persze mindezt megoldhattuk volna egyszerűbben."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class Person(object):\n",
    "    def __init__(self, name, title=\"\"):\n",
    "        self.name = name\n",
    "        self.title = title\n",
    "    def __repr__(self):\n",
    "        return self.title + \" \" + self.name\n",
    "\n",
    "class Knight(Person):\n",
    "    title = 'Sir'\n",
    "    searching_for = 'The Holy Grail'\n",
    "    def __init__(self, name, eo):\n",
    "        super(Knight, self).__init__(name)\n",
    "        self.eo = eo\n",
    "    def __repr__(self):\n",
    "        return Knight.title + \" \" + self.name + \", \" + self.eo    \n",
    "\n",
    "class Group:\n",
    "    def __init__(self, name, persons):\n",
    "        self.persons = persons\n",
    "        self.name = name\n",
    "    def __iter__(self):\n",
    "        return iter(self.persons)\n",
    "        \n",
    "kotrt = Group('Knights of The Round Table', \n",
    "              [Knight('Launcelot', 'the brave'), \n",
    "               Knight('Galahad', 'the pure'),\n",
    "               Knight('Bedevere', 'the wise'), \n",
    "               Knight('Robin', 'the Not-quite-so-brave-as-Sir-Launcelot')])\n",
    "for knight in kotrt:\n",
    "    print knight"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "source": [
    "# Függvények argumentumai\n",
    "Lehetőségünk van függvényeket, illetve metódusokat többféle módon meghívni. Ismerjük, hogy hogyan lehet default paraméterértéket megadni."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "class A(object):\n",
    "    def __init__(self):\n",
    "        pass\n",
    "    def fuggveny(self):\n",
    "        print \"Eredeti fuggveny\"\n",
    "\n",
    "class B(A):\n",
    "    def fuggveny(self, i):\n",
    "        print \"B osztaly fuggvenye es\", i\n",
    "        \n",
    "class C(A):\n",
    "    def fuggveny(self, i=None):\n",
    "        if i is None:\n",
    "            super(C, self).fuggveny()\n",
    "        else:\n",
    "            print \"C osztaly fuggvenye es\", i\n",
    "        \n",
    "a = A()\n",
    "a.fuggveny()\n",
    "\n",
    "b = B()\n",
    "b.fuggveny(2)\n",
    "\n",
    "c1 = C()\n",
    "c1.fuggveny()\n",
    "\n",
    "c2 = C()\n",
    "c2.fuggveny(3)"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Írjunk függvényt, ami felveszi a rendelést! Ha nem említjük, hogy mennyit kérünk, akkor úgy érti, hogy egyet."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "def order(name, quantity=1):\n",
    "    print name, quantity\n",
    "order('eggs', 2)\n",
    "order('spam')"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Most megoldjuk, hogy bármennyi különböző terméket rendelhessünk. Használjuk a `*argumentumok` formátumú argumentumokat."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "def order(name, *args):\n",
    "    print name\n",
    "    for a in args:\n",
    "        print a\n",
    "order('sausage', 'bacon', 'egg', 'spam')        "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Ha különböző mennyiségeket akarunk rendelni, akkor másik módszerhez kell folyamodni: használjuk a `**argumentumok` formátumú argumentumokat."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "def order(*args, **kwargs):\n",
    "    for k in kwargs:\n",
    "        print k, kwargs[k]\n",
    "order('egg', 'sausage', spam=4)        "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "# Röviden a kivételekről\n",
    "Vannak olyan esetek, amikor a kód hibába ütközik. Ilyenkor egy Exception típusú **objektumot** dob a python. Ha senki nem kapja el, akkor a kód megáll futni. Egy Exceptiont bármikor el lehet kapni, akármennyi fügvényhívást túlél. Nézzünk egy példát! Most mi magunk hívjuk elő a hibát, de egyben le is kezeljük."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "try:\n",
    "   raise Exception('spam', 'eggs')\n",
    "except Exception as inst:\n",
    "   print type(inst)    \n",
    "   print inst.args      \n",
    "   print inst           \n",
    "   x, y = inst.args\n",
    "   print 'x =', x\n",
    "   print 'y =', y"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Most pedig lekezelünk egy nem mesterségesen létrehozott kivételt."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "try:\n",
    "   str(5) + 5\n",
    "except TypeError as inst:\n",
    "   print type(inst)\n",
    "   print inst.args     \n",
    "   print inst           "
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {
    "deletable": true,
    "editable": true
   },
   "source": [
    "Természetes környezetben. Ha a felhasználó nem egész számot ad, nem tudja azzá konvertálni, így hibát jelez, mi viszont ezt lekezeljük."
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": false,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": [
    "while True:\n",
    "    try:\n",
    "        x = int(raw_input(\"Please enter a number: \"))\n",
    "        break\n",
    "    except ValueError:\n",
    "        print \"Oops!  That was no valid number.  Try again...\"\n",
    "print 2*x"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": null,
   "metadata": {
    "collapsed": true,
    "deletable": true,
    "editable": true
   },
   "outputs": [],
   "source": []
  }
 ],
 "metadata": {
  "kernelspec": {
   "display_name": "Python 2",
   "language": "python",
   "name": "python2"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 2
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython2",
   "version": "2.7.12"
  }
 },
 "nbformat": 4,
 "nbformat_minor": 0
}
